/*
 * Mach Operating System
 * Copyright (c) 1992, 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * Handles the Amiga's internal RS232 port
 * *very* losely based on the luna88k code...
 *
 * Contributors:
 *      Markus Wild
 *      Bryan Ford
 */


#include <sys/types.h>

#include <kern/time_out.h>
#include <device/conf.h>
#include <device/tty.h>
#include <device/io_req.h>
#include <device/buf.h>         /* for struct uio (!) */
#include <device/errno.h>

#include <sys/ioctl.h>

#include <machine/machparam.h>
#include <machine/uart.h>
#include <machine/custom.h>

extern void ttrstrt();

struct tty uart_tty = {};

/*
 * map speed numbers into the values we have to poke into SERPER
 * num = int((1 / (baudrate * 2.79365e-7)) - 1)
 */
short   uart_speeds[16] = {  /* XXX - Do we need separate PAL/NTSC versions? */
/* B0   */      0,
/* B50  */      0,
/* B75  */      0,
/* B110 */      32540,
/* B134 */      26712,
/* B150 */      23863,
/* B200 */      17896,
/* B300 */      11931,
/* B600 */      5965,
/* B1200*/      2982,
/* B1800*/      1988,
/* B2400*/      1490,
/* B4800*/      745,
/* B9600*/      372,
/* EXTA */      185,
/* EXTB */      92,
    };

#ifndef TOSTOP
#define TOSTOP TS_TTSTOP
#endif

#define IFLAGS (TOSTOP|PASS8|EVENP|ODDP|ECHO|XTABS|CRMOD)


static int
utmodem(unsigned dev, int flag)
{
#if 0
    if (flag == OFF)
        utlsi(line,WR5,~(DTR|RTS),OFF);
    else
        utlsi(line,WR5,(DTR|RTS),ON);
    return(uartaddr->status & CD);
#endif
    return 1;
}


static void
utstop (struct tty *tp, int flag)
{
  unsigned s;

  s = spltty();
  if (tp->t_state & TS_BUSY)
      if ((tp->t_state&TS_TTSTOP)==0)
          tp->t_state |= TS_FLUSH;
  splx(s);
}


static int
utstart(struct tty *tp)
{
  int c, x;

  x = spltty();
  if (tp->t_state & (TS_TIMEOUT|TS_BUSY|TS_TTSTOP))
      goto out;

  if (tp->t_outq.c_cc <= TTLOWAT(tp)) {
          tt_write_wakeup (tp);
  }

  if (tp->t_outq.c_cc == 0)
      goto out;

  tp->t_state |= TS_BUSY;
  c = getc (&tp->t_outq);
  if (tp->t_flags & (RAW|LITOUT|PASS8))
    {
      custom.serdat = (c&0xff) | 0x100;
    }
  else if (c <= 0177)
    {
      custom.serdat = (c&0xff) | 0x100;
    }
  else
    {
      tp->t_state &= ~TS_BUSY;
      timeout((int(*)())ttrstrt, (caddr_t) tp, c&0xff);
      tp->t_state |= TS_TIMEOUT;
    }
 out:
  splx(x);
  return(1);
}

static void
utparam(unsigned dev)
{
  struct tty *tp;
  unsigned short serper;
  unsigned s;

  tp = &uart_tty;

  /* hang up line */
  if (tp->t_ispeed == 0) {
      utmodem(dev, OFF);
      return;
  }
  s = spltty();

  /* for now, only look at ispeed... */
  serper = uart_speeds[tp->t_ispeed];
  /* for now, no 2 stop bits are supported, neither is parity.. */
  custom.serper = serper;

  splx(s);
}

int
utopen(unsigned dev, int flag, io_req_t ior)
{
  struct tty *tp;
  unsigned s;

  tp = &uart_tty;

  tp->t_stop = (int (*)())utstop;

  s = spltty();
  tp->t_oproc = (int (*)())utstart;
  if ((tp->t_state&TS_ISOPEN) == 0) {
      ttychars(tp);
      tp->t_flags = IFLAGS;
      tp->t_ospeed = tp->t_ispeed = B9600;
      utparam(dev);
  }

  if(utmodem(dev, ON))
    ttymodem(tp, 1);

  splx(s);
  return char_open(dev, tp, flag, ior);
}

int
utclose(unsigned dev)
{
  struct tty *tp;
  unsigned  s;

  tp = &uart_tty;
  if (tp->t_state&TS_HUPCLS && !(tp->t_state&TS_ISOPEN))
      utmodem(dev, OFF);
  s = spltty();
  ttyclose(tp);
  splx(s);
  return 0;
}

int
utread(unsigned dev, io_req_t ior)
{
    struct tty *tp;

    tp = &uart_tty;
    return char_read(tp, ior);
}

int
utwrite(unsigned dev,io_req_t ior)
{
  struct tty *tp;

  tp = &uart_tty;
  return char_write(tp, ior);
}

io_return_t
utgetstat (unsigned dev, int flavor, int *data, unsigned count)
{
  return tty_get_status (&uart_tty, flavor, data, count);
}


io_return_t
utsetstat (unsigned dev, int flavor, int *data, unsigned count)
{
  return tty_set_status (&uart_tty, flavor, data, count);
}


void
utrint ()
{
  struct tty *tp;
  unsigned short data;
  unsigned char c;

  tp = &uart_tty;

  data = custom.serdatr;
  /* acknowledge that we received the character */
  custom.intreq = INTF_RBF;

  if (!(data & SERDATRF_RBF))
    return;

  if ((tp->t_state&(TS_ISOPEN|TS_WOPEN)) == 0)
      return;

  if (tp->t_flags&(RAW|LITOUT|PASS8)) /*CS8*/
    c = data & 0xff;
  else                                /*CS7*/
    c = data & 0x7f;

  ttyinput(c, tp);
}

void
utxint ()
{
  struct tty *tp;

  tp = &uart_tty;

  /* acknowledge that we know the UART is no longer busy */
  custom.intreq = INTF_TBE;

  if((tp->t_state & (TS_ISOPEN|TS_WOPEN))==0)
    return;

  if (tp->t_state & TS_FLUSH)
    tp->t_state &= ~TS_FLUSH;
  if (tp->t_state & TS_BUSY)
    tp->t_state &= ~TS_BUSY;

  utstart (tp);
}

void
utinit()
{
  unsigned s = spltty();

  /* default.. might want to fiddle with the CIA here later */
  custom.serper = uart_speeds[B9600];

  splx (s);
}

int
utportdeath(unsigned dev, mach_port_t port)
{
  return (tty_portdeath(&uart_tty, (ipc_port_t)port));
}


/* Called from cons.c, to output characters by polling.  */
void
ut_cnputc(int c)
{
  if (c)
    {
      /* inhibit tx-interrupts while we poll manually, or the uart driver
         may get `slightly' confused.. */
/*      custom.intena = INTF_TBE; */
      while (! (custom.serdatr & (SERDATRF_TBE|SERDATRF_TSRE))) ;
      custom.serdat = (c&0xff) | 0x100;
      while (! (custom.serdatr & SERDATRF_TBE)) ;
      custom.intreq = INTF_TBE;
/*      custom.intena = INTF_SETCLR | INTF_TBE; */

#if 1
      /* well ahem... *slow* vt100.... */
      for (c = 0; c < 10000; c++) ;
      
      /* rather rude kind of flow-control, but I want to see EVERYTHING on
         my vt100, not just parts of a too fast transmission ;-)) */
      if ((c = ut_cnpollc ()) == 19)
        {
          do c = ut_cnpollc (); while (c != 17);
	}
#endif
    }
}


/* Called from cons.c, to poll for a character.  */
int
ut_cnpollc(void)
{
    register int s,c;
   
/*    custom.intena = INTF_RBF; */
    c = (custom.serdatr) & 0xffff;
    if (! (c & SERDATRF_RBF))
      c = -1;
    else
      {
        c &= 0x7f;
        custom.intreq = INTF_RBF;
      }
/*    custom.intena = INTF_SETCLR | INTF_RBF; */

#if MACH_KDB
    if ( c == ddb_trap_char)
      {
	c = -1;
	gimmeabreak();
      }
#endif
    return c;
}

